// 从 upx 拉取模块信息

const _ = require('lodash');
const request = require('sync-request');
var semver = require('semver');
const path = require('path');
var fs = require('fs');
var Promise = require('bluebird');

var regroup = require('./lib/regroup');
var util = require('./lib/util');


var Debug = require('debug');
var debug = Debug('deps');
var debugMini = Debug('deps:mini');
var debugError = Debug('deps:error');
var debugExtra = Debug('deps:extra');

/**
 * 从 upx 中获取所需要的依赖关系
 * @param  {string} mainEntryModName 入口名，比如 'upx/@ali/tb-banner-1212/index'
 * @param  {json} dependencies 语义化依赖描述，一般从 package.json 中的 dependencies 中直接获取
 * @param  {[type]} comBasePath      组件基准路径，默认是 'npm'
 * @param  {[type]} suffix           后缀，默认是 '.cmd'
 * @return {json}    检索获取的依赖对象，如果请求失败抛出错误
 */
function getDepsFromUpx(mainEntryModName, dependencies, comBasePath, suffix){
	
	// 如果是空的，制造标准的空白格式
	if(_.isEmpty(dependencies)){
		// 直接返回空白依赖，不需要查询 upx
		return new Promise.resolve({
			depMap: genEmptyDepMap(mainEntryModName),
			verJSON: {}
		});
	}


	/* ----------------------------------------------------
	    	如果有依赖，需要去 upx 查找
	----------------------------------------------------- */
	return util.upx.request('npmService.getNpmVersionBySemVersion',[dependencies]).then(function(result) {

		// var result = util.parseOrFalse(data);
		debugExtra('======> 从 upx 检索获得的 deps：', result);

		if(!result){
			// console.log('从 upx 检索依赖的数据格式有误，解析json出错');
			throw new Error('从 upx 检索依赖的数据格式有误，解析json出错');
		}

	    return genDepsFromResponse(result, mainEntryModName, comBasePath, suffix);

	}).catch(function(err){
		console.log('处理请求发生错误:', err);
		throw new Error('从 upx 检索依赖失败，请检查网络状况 或 返回的数据格式有误');
	});
	
}


/**
 * 根据 upx 返回的数据结果生成 deps.json 文件
 * @param  {json} result  从 upx 返回的数据结果
 * @return {json}         生成的 deps.json 文件
 */
function genDepsFromResponse(result, mainEntryModName, comBasePath, suffix) {
	var resultDeps = {}; // 存储最终的deps.json
	var verJSON = {}; // 存储版本号信息
	var depArr = [];

	// 如果存在 result
	if(result){
		
		// 组装成所需要的格式
		// key : @ali/rx
		// value : { version: 0.1.21 , depsJsonStr: 'xxxx'}
		_.forOwn(result, function(value, key) {

			verJSON[key] = value && value.version;

			if(value !== null && value !== 'null'){

				var name = [comBasePath, key, value.version,'index'+suffix].join('/');
				
				debugMini('======> deps组件名：', name);

				// 放入到入口文件的依赖数组
				depArr.push(name);

				// TODO: 如果 depsJsonStr 为 null 的情况下，是否需要创建空白对象
				var currentDepMap = util.parseOrFalse(value.depsJsonStr) || {};

				// 查询获得的 depmap 是没有后缀的，需要处理一番；
				// 过滤不需要的模块，比如 @weex-module 等等
				var depMapWithSuffix = filterAndSuffix(name, currentDepMap, suffix); 

				// 将其归并到 resultDeps 对象
				_.assign(resultDeps, depMapWithSuffix);
			}


		});

		// 生成主内容
		resultDeps[mainEntryModName] = {
			requires: depArr
		}

	}

	// 生成一份 version

	return {
		depMap: resultDeps,
		verJSON: verJSON
	} // 返回所需要的依赖对象
}


// 生成空白depMap
function genEmptyDepMap(modName) {
	var depMap = {};

	depMap[modName] = {
		requires: []
	}

	return depMap;
}

// 处理依赖的后缀
// 同时过滤掉不需要的模块
// { a/index : requires:['b/index']} -> { a/index.cmd : requires:['b/index.cmd']}
function filterAndSuffix(modName, depMap, suffix) {

	if(_.isEmpty(depMap)){ // 做一下兼容，不存在的就生成空白
		return genEmptyDepMap(modName);
	}

	var newDepMap = {};

	_.forOwn(depMap, function(value, key){

		var depArr = value && value.requires || {};

		// 注意，有些情况是不能添加后缀的，比如以 .xtpl / .cmd 后缀的
		var canSuffix = util.canXtplSuffix(key) && util.canCmdSuffix(key);

		if(depArr){
			// 过滤模块
			depArr = util.filterIgnore(depArr, [], 'first');
			
			// 添加后缀
			if(canSuffix){
				newDepMap[key+suffix] = {
					requires: depArr.map(function(value){
							return value+suffix;
					})
				};
			} else {
				newDepMap[key] = {
					requires: depArr
				}; 
			}
		}
	});

	debugMini('=====> 经过过滤&后缀处理的依赖表：', newDepMap);
	return newDepMap;
}




// 如果文件
module.exports = function(options) {

	debugMini('=====> 依赖入参：', options);

	// 从 options 中解构参数
	var dependencies = options.dependencies || {};
	var type = options.type; 
	var modName = options.modName; // 模块名字 
	var suffix = options.suffix;
	var entrySuffix = options.entrySuffix || '';
	var comBasePath = options.comBasePath;
	var modBasePath = options.modBasePath;
	var disableBuiltIn = options.disableBuiltIn || false;
	var version = options.version; // 是否存在版本号（组件是有版本号的）

	var disableXtplReplace = options.disableXtplReplace;
	var disableBuiltIn = options.disableBuiltIn;


	// 拼接获取入口
	var entryArr = version ? [modBasePath, modName, version, 'index'] : [modBasePath, modName, 'index'];
	var mainEntryModName = entryArr.join('/');

	// 获取最初始的依赖， 从 upx 那儿获取所需的依赖
	return getDepsFromUpx(mainEntryModName, dependencies, comBasePath, suffix).then(function(result){

		var depMap = result.depMap;

		debugExtra('=====> 最原始的完整依赖：', JSON.stringify(depMap));

		debugExtra('=====> 获取的版本信息：', JSON.stringify(result.verJSON));

	    // 根据类型，获取新的 depMap
	    var newDepMap = regroup({
	      disableXtplReplace: disableXtplReplace,
	      disableBuiltIn: disableBuiltIn,
	      type: type, // 终端类型
	      suffix: suffix, // 后缀
	      entrySuffix: entrySuffix, // 主入口后缀
	      mainEntryModName: mainEntryModName, // 主入口名
	      depMap: _.assign({}, depMap) // 原始依赖，拷贝一份，防止原始依赖被篡改
	    });

	    debugExtra('=====> 重组之后的依赖：', JSON.stringify(newDepMap));

	    // 合并依赖表
	    depMap = _.merge({}, depMap, newDepMap);

	    // 如果设置了 entrySuffix ,需要删除原始主入口，不要使用 delete 操作符
	    if(entrySuffix){
		    depMap[mainEntryModName] = undefined;
	    }
	    
	    debugExtra('=====> 合并之后的依赖：', JSON.stringify(depMap));

		return {
			depMap: depMap,
			verJSON: result.verJSON
		}
	});
    
};
