var fs = require('fs');
var path = require('path');
var _ = require('lodash');
var request = require('sync-request');
var Promise = require('bluebird');
var Client = require('isv-module-info');


var Debug = require('debug');
var debug = Debug('deps');
var debugMini = Debug('deps:mini');
var debugExtra = Debug('deps:extra');
var debugError = Debug('deps:error');

var config = require('../config');

var upx = Client.upxClient(config.upxEnv);


// 正则表达式
var regWeb = /\.web(\.js)?$/;
var regNative = /\.native(\.js)?$/;
var regComLib = /kg\/rx\-components\/[\d\.]+\/lib/;

var xtplSuffix = '.xtpl';
var cmdSuffix = '.cmd';
var regXtplSuffix = xtplSuffix ? new RegExp('\\'+xtplSuffix+'$') : false;
var regCmdSuffix = cmdSuffix ? new RegExp('\\'+cmdSuffix+'$') : false;


// 根据entry名字，决定是否可以添加xtpl后缀
function canXtplSuffix(entry){

  // 以 .xtpl 结尾的，不需添加后缀
  if(regXtplSuffix && regXtplSuffix.test(entry)){
    return false;
  }

  return true;
}

// 根据entry名字，决定是否可以添加 cmd 后缀
function canCmdSuffix(entry) {

  // 以 .cmd 结尾的，不需要添加后缀了
  if(regCmdSuffix && regCmdSuffix.test(entry)) {
    return false;
  }
  return true;
}



function readFileOrEmpty(path) {
  try {
  	debugMini('=====> 读取 '+ path + '文件内容');
    return fs.readFileSync(path,'utf8');
  } catch(e) {
  	debugError('=====> 读取 '+ path + '文件内容失败',e);
    return "";
  }
}

// 将数据写入文件
function writeFileOrNone(path, data) {

  fs.writeFile(path, data, 'utf8',function(err){
    if(err){
      console.log('=====> 写入数据到'+path+'失败');
      debugError('=====> 想要写入的数据：'+data,err);
      return;
    }

    debugMini('=====> 已将数据写入到：'+path);
    debug('=====> 文件内容：'+data);
  });
}


// 将数据写入文件
function writeFileSyncOrNone(path, data) {
  
  debug('=====> 要写入的文件内容：',data);

  try {
    fs.writeFileSync(path, data);
    debugMini('=====> 写入文件成功：' + path);
  } catch(e) {
     debugError('=====> 写入失败：',e);
     console.log('文件',path,'写入失败');
  }

}



// 新建文件夹
function mkdirSyncOrNone(path){
  try {
    fs.mkdirSync(path);
    debugMini('=====> 创建文件夹成功：' + path);
  } catch(e) {
    if ( e.code == 'EEXIST' ){
      debugMini('=====> 文件夹已存在不需要创建');
    } else { // 如果不是文件已存在的案例，则直接抛出错误
      debugError('=====> 创建文件夹失败：',e);
      console.log('文件夹:',path,'创建失败');
      throw e; 
    }
  }
}



// 计算同步执行callback所花费的时间
function calcTime(fn,msg){
  var oStart = new Date();
  fn();
  var oEnd = new Date();
  console.log(msg,':',(oEnd-oStart)/1000,'s');
}

// 解析字符串
function parseOrFalse(str){
  var json = {}
  try{
    if(str.length){
      json = JSON.parse(str);
      debugMini('=====> 将字符串解析成 JSON 成功');
      debug('=====> 获得的 JSON 的内容：'+json);
      return json;
    } else {
      return false;
    }
  } catch(e) {
    debugError('=====> 将字符串解析成 JSON 失败：'+e);
    debugError('=====> 原字符串内容：'+str);

    return false;
  }
}


// 获取终端类型
// 入参：entry名字，比如 'tb-countdown-1111/index.web'
// 出参：枚举，origin, web, native
function getEndType(entryName){

  var type = 'origin';

  // 容错
  if(typeof entryName === 'undefined' ||  entryName === null) {
    return type;
  }

  if(regWeb.test(entryName)){
    type = 'web';
  } else if(regNative.test(entryName)){
    type = 'native';
  }

  debugExtra('=====> 终端类型：', type);

  return type;
}


// 依据终端类型返回 deps 文件名，默认返回 deps.json
function getDepsFileName(type){

  var arr = [];

  arr.push(config.depsFileName);

  if(typeof type !== 'undefined' &&  type !== null){
    arr.push(type);
  }

  arr.push('json');

  return arr.join('.');
  
}

// 依据终端类型返回 version 文件名，默认返回 version.json
function getVerFileName(type){
  var arr = [];

  arr.push(config.versionFileName);

  if(typeof type !== 'undefined' &&  type !== null){
    arr.push(type);
  }

  arr.push('json');

  return arr.join('.');
}

// 根据 ignoreList 过滤指定deps
// flag 是为了标志是哪段逻辑中执行过滤操作的，debug用
function filterIgnore(arr, ignoreList, flag){

    return  _.filter(arr, function(dep){

        if(ignoreList.indexOf(dep) !== -1){
          debugMini('=====> ('+flag+')过滤模块：' + dep + ', 不放入deps.json');
          return false;
        }

        // 需要去除.css模块
        if(/\.css$/.test(dep)){
          debugMini('=====> ('+flag+')过滤css模块：' + dep + ', 不放入deps.json');
          return false;
        }

        // 需要去除weex模块
        if(/@weex-module/.test(dep)){
          debugMini('=====> ('+flag+')过滤weex模块：' + dep + ', 不放入deps.json');
          return false;
        }

        return true;
    });

}

// 客户端内置的白名单模块
function getBuiltInMods() {
  // 配置地址: http://ems.alibaba-inc.com/fragment/technical/maintain/daily/global/rx/external/list
  var res = request('GET', 'http://www.taobao.com/go/rgn/ems/global/rx/external/list');
  res = JSON.parse(res.getBody());
  var list = res['list'].map(function(item){
    var info = item.split('/');
    return 'npm/@ali/'+info[0];
  });

  // 需要对rx-components做特殊处理
  // _.remove(list, function(mod){
  //   return mod === 'kg/rx-components';
  // });

  return list;
}

// 从 upx 获取版本信息
function getVersionFromUpx(dependencies) {
  
  // 如果是空的，制造标准的空白格式
  if(_.isEmpty(dependencies)){
    // 直接返回空白依赖，不需要查询 upx
    return new Promise.resolve({});
  }

  /* ----------------------------------------------------
        如果有依赖，需要去 upx 查找
  ----------------------------------------------------- */
  return upx.request('npmService.getNpmVersionBySemVersion',[dependencies]).then(function(result) {

    // var result = util.parseOrFalse(data);
    debugExtra('======> 从 upx 检索获得的 deps：', result);

    if(!result){
      // console.log('从 upx 检索依赖的数据格式有误，解析json出错');
      throw new Error('从 upx 检索依赖的数据格式有误，解析json出错');
    }

    // 组装成 version 格式
    var verJSON = {};
    _.forOwn(result, function(value, key) {

      verJSON[key] = value && value.version;

    });

    return verJSON;

  }).catch(function(err){
    console.log('处理请求发生错误:', err);
    throw new Error('从 upx 检索依赖失败，请检查网络状况 或 返回的数据格式有误');
  });
}

// 确保缓存文件夹存在
function checkCacheDirExist(pwd) {
    var CACHE_DIR = path.join(pwd, config.cacheDir); // 缓存文件夹
    if(!fs.existsSync(CACHE_DIR)){ 
      mkdirSyncOrNone(CACHE_DIR); 
    }
}

module.exports = {
  upx: upx,
	readFileOrEmpty : readFileOrEmpty,
	writeFileOrNone : writeFileOrNone,
  writeFileSyncOrNone: writeFileSyncOrNone,
	calcTime: calcTime,
  mkdirSyncOrNone: mkdirSyncOrNone,
  getEndType: getEndType,
  getDepsFileName: getDepsFileName,
  getVerFileName: getVerFileName,
  filterIgnore: filterIgnore,
  getBuiltInMods: getBuiltInMods,
  canXtplSuffix: canXtplSuffix,
  canCmdSuffix: canCmdSuffix,
  getVersionFromUpx: getVersionFromUpx,
  checkCacheDirExist: checkCacheDirExist,
  parseOrFalse: parseOrFalse
}