/**
 * 用于抽取模块依赖表的 webpack 插件
 * @param options
 * @constructor
 */
var fs = require('fs');
var _ = require('lodash');
var path = require('path');
var Promise = require('bluebird');


var Debug = require('debug');
var debug = Debug('deps');
var debugMini = Debug('deps:mini');
var debugError = Debug('deps:error');
var debugExtra = Debug('deps:extra');

var getDepMap = require('./index');
var config = require('./config');

var util = require('./lib/util');

var PATH_MOD_BASE = 'upx'; // 模块的基准路径
var PATH_COM_BASE = 'npm'; // 组件的基准路径


// 匹配define的入口名，要求不以模块基准路径开始，使用正向否定断言
var REG_ENTRY = new RegExp('define\\(\\"(?!'+PATH_MOD_BASE+'\\/)(\\S*?)\\",');


// 匹配define部分，比如 define("rx-demo/index",["@ali/rx/index","@ali/rx-slider/index"],function(e,t,n
var REG_DEPS = /define\(\".+\",\s?(\[\S*?\]),\s?function\(/; 

// 匹配 kg 组件，比如 "kg/rx-components/index"、"kg/rx/index"
// var REG_KG_COM = /"kg\/([\/\-\w]+)"/g; 

// 匹配 require 部分，比如 require("isv-rx-components/index")，使用正向否定断言
// var REG_REQUIRE = new RegExp('require\\("(?!'+PATH_COM_BASE+'\\/)(\\S+?)"\\)','g');

var cacheUpxVersion = {}; // 用于缓存 upx 版本

function isExist(obj){
  return typeof obj !== 'undefined' &&  obj !== null;
}

// 替换内容
// 1. define("a") -> define("upx/tb-mod/a")
function replaceContent(source, group, modBasePath, npmBasePath, isComponent){
  return function(){

    // 如果是组件模块，直接返回
    if(!isComponent){
      source = source.replace(REG_ENTRY,'define("'+modBasePath+'/'+group+'/\$1",');
    }

    return source;

  }
}

// 将从 upx 那儿返回过来的版本号缓存到 version.upx.json 中
function writeUpxVersion(compiler, isDev, isNewModule, cache,  config, callback){

    var entries = [];

    // 只在 dev 情况下才会将 version.upx.json 输出到本地
    if(!isDev){
       return callback();
    }

    // compiler.context: light/mod/
    // compiler.options.entry: {'test4-xn-mod/index': xxx, 'test4-xn-mod/index.web': xxx}
    _.forOwn(compiler.options.entry, function(value, key){
      var info = key.split('/');
      var entry = path.join(compiler.context,info[0]);
      
      // 缓存之前的入口，避免重复读取文件，只在 dev 下有效果
      if(!cache[entry]){
        entries.push(entry);
      }
    });

    entries = _.uniq(entries); // 去重
    debugExtra('=====> 去重之后的入口数组：',entries);
  

    Promise.all(entries.map(function(entry){
      
      var pwd = entry;
      // 读取packageJSON
      var packageJSON = util.parseOrFalse(util.readFileOrEmpty(path.join(pwd, 'package.json')));


      var guideName = packageJSON.upx && packageJSON.upx.guideName; // 是否是新模块（基于upx查询）
      
      if(!guideName && !isNewModule) return; // 不是新模块就直接退出

      return util.getVersionFromUpx(packageJSON.dependencies).then(function(result){
        
        // 缓存文件夹，先创建缓存文件夹，做一层容错
        util.checkCacheDirExist(pwd);

        // 生成对应端的 version.json 文件，放到缓存文件中去
        var verFile = path.join(pwd, config.cacheDir, 'version.upx.json');
        
        // 过滤结果，给出提示
        _.forOwn(result, function(v, k){
          if(typeof v === 'undefined' ||  v === null){
            console.log(' ****** 组件 ' + k + ' 未在 upx 中注册，若引用该组件将打包进最终bundle，请斟酌使用 *****\n');
          }
        });

        // 输出版本
        util.writeFileSyncOrNone(verFile, JSON.stringify(result));

        // 在dev下使用缓存功能
        cache[pwd] = true; 

      });

    })).then(function(){
      callback();
    }).catch(function(err){
      console.log('从upx获取的版本写入到 version.upx.json 失败',err);
      callback();
    });

}

// 依赖替换成源码中真正使用的
function replaceTrulyDeps(modDeps, entryNames, trulyDeps, verJSON){

  // 如果没有entryNames，就不更改
  if(!entryNames || !entryNames.length){
    return modDeps;
  }

  var newModDeps = _.cloneDeep(modDeps); // 先拷贝一份
  
  entryNames = [].concat(entryNames); // 变成数组

  _.forEach(entryNames, function(entryName){

    var baseName = path.basename(entryName); // baseName: index.xtpl 、 index.native.xtpl

    var baseNameInfo = baseName.split('.');

    var isXtpl = _.indexOf(baseNameInfo, 'xtpl') > 0; // xtpl 应该出现位置不在第一个
    var isWeb = _.indexOf(baseNameInfo, 'web') > 0;
    var isNative = _.indexOf(baseNameInfo, 'native') > 0;

    // 文件后缀名的区分
    var backupSuffix = isXtpl ? '.xtpl' : '.cmd';

    var fileSuffix = '.cmd';
    if(isXtpl && !isNative){
      fileSuffix = '.xtpl';
    } else if(isXtpl && isNative) {
      fileSuffix = '.native.xtpl';
    } else if(isWeb){
      fileSuffix = '.web.cmd';
    } else if(isNative) {
      fileSuffix = '.native.cmd';
    }

    var curDeps = newModDeps[entryName];
    // 将真正依赖注入到依赖中
    if(curDeps){
      var newDepsArr = [];

      _.forEach(trulyDeps, function(v){
        var info = v.split('/'); // v: npm/@ali/pi-kissy-pc/mod-base
        var insertPos; // 插入版本的位置

        if(info[0]==='npm'){
          curModName = [info[1], info[2]].join('/'); // curModName: @ali/pi-kissy-pc
          insertPos = 3;
        } else {
          curModName = info[0]; // 获取真实组件名，默认是第一个，比如社区组件 rax
          insertPos = 2;
          
          info.unshift('npm'); // 变成 ['npm', 'rax']
          if(info.length == 2){
            info.push('index'); // 补全成 ['npm', 'rax', 'index']
          }
        }

        var curVer = verJSON[curModName];

        // 对 xtemplate 做特殊版本号处理
        if(info[2] === 'xtemplate'){
          curVer = '4.5.0';
        }

        // 如果存在版本号
        if(curVer){
          info.splice(insertPos, 0, curVer);

          // 查看对应的key是否存在于modDeps，如果存在则使用当前的
          var targetPath = info.join('/')+fileSuffix;

          if(!newModDeps[targetPath]){
            targetPath = info.join('/') + backupSuffix;
          }

          newDepsArr.push(targetPath);
        }

      });

      // npm/@ali/pi-kissy-pc/0.0.5/mod-base
      newModDeps[entryName] = {requires: newDepsArr};
    }
  });

  return newModDeps;
}


function ExternalDeps(options) {
  this.options = options;
}

module.exports = ExternalDeps;

ExternalDeps.prototype.apply = function(compiler) {
  
  var options = this.options;
  var isDev = options.isDev;
  var isComponent = options.isComponent; // 是否是组件
  var depsJsonDest = options.depsJsonDest; // 指定输出目录
  var isNewModule = options.isNewModule; // 是否强制指定是用新upx规范

  //入口目录
  var entryDir = options.options.entryDir;

  // 获取chunk的base目录，优先取context（绝对路径）
  var basePath = options.options.context || entryDir;


  // 从 upx 获取版本信息
  compiler.plugin('run', function(compiler, callback) {
    writeUpxVersion(compiler, isDev, isNewModule, cacheUpxVersion, config, callback);
  });

  // watch 的情况
  compiler.plugin('watch-run', function(watching, callback){
    var compiler = watching.compiler;
    writeUpxVersion(compiler, isDev, isNewModule, cacheUpxVersion, config, callback);
  });

  compiler.plugin('emit',function(Compilation, callback){
    var assets = Compilation.assets;
    var shouldRecreateDeps = true;

    if (isDev) {
      // 如果 deps.json 存在并且 dependencies 没有发生变更，则不重新生成
      var outputDir = path.join(Compilation.outputOptions.path, path.basename(basePath));
      var prevPackageInfo;

      try {
        prevPackageInfo = JSON.parse(fs.readFileSync(path.join(outputDir, 'package.json')));

        if (fs.existsSync(path.join(basePath, 'deps.json')) && !prevPackageInfo.changed) {
          process.utils.log.verbose('依赖没有变更，不再重新生成 deps.json');
          // 之前依赖的重新更新在 builder-mod 里已经有处理
          shouldRecreateDeps = false;
        }
      } catch (e) {
        process.utils.log.verbose('获取之前缓存的 dependencies 失败，重新生成');
      }
    }

    // 迭代所有的 assets
    Promise.all(_.keys(assets).map(function(entry){

        if(!/\.js$/.test(entry)) return;

        //window 兼容
        entry = entry.replace('\\','/');

        // 如果是 dev，entry 是带模块名的，比如 'test4-xn-mod/index.js'
        // 而在真正打包的时候，一般直接是 ‘index.js’
        var pwd = basePath;
        //if(isDev){
        //  var info = entry.split('/');
        //  pwd = path.join(basePath, info[0]);
        //}
        var packageJSON = util.parseOrFalse(util.readFileOrEmpty(path.join(pwd, 'package.json')));
        var version = packageJSON.version;

        var guideName = packageJSON.upx && packageJSON.upx.guideName; // 是否是新模块（基于upx查询）
        // var isRaxMod = abcJSON.info && abcJSON.info.extraLib === 'rax'; // rax 模块
        
        if(!guideName && !isNewModule) return; // 不是新模块就直接退出


        var abcJSON = util.parseOrFalse(util.readFileOrEmpty(path.join(pwd, 'abc.json')));
        var group = abcJSON.group;

        // 缓存文件夹，先创建缓存文件夹，做一层容错
        util.checkCacheDirExist(pwd); 

        /* ----------------------------------------------------
           rax 源码替换，添加 upx/ npm/ 前缀
        ----------------------------------------------------- */
        // 获取 source
        var source = assets[entry].source();

        // rax 的替换规则
        Compilation.assets[entry].source = replaceContent.call(this, source, group, PATH_MOD_BASE, PATH_COM_BASE, isComponent);

        // === end ===

        if (!shouldRecreateDeps) {
          return;
        }

        
        // 根据组件取用不同的group
        var modName = isComponent ? packageJSON.name : [group ,abcJSON.name].join('/'); 
        var endType = util.getEndType(entry);
        var dependencies = packageJSON.dependencies;

        // 从依赖表中获取所需要的 dependencies
        var resultSource = assets[entry].source();
        var matches = resultSource.match(REG_DEPS);
        var matchDeps = matches && matches[1] || '[]';
        var trulyDeps = util.parseOrFalse(matchDeps); // 从依赖表中获取所需的 deps

        debugExtra('=====> 关键入参：','dependencies:',dependencies,'; group:',options.group, '; modName:', modName, '; endType:',endType, '; isDev:', options.isDev);

        // 默认给 模块 使用的配置
        var depConfig = {
          dependencies: dependencies,
          modBasePath: PATH_MOD_BASE,
          comBasePath: PATH_COM_BASE,
          modName: modName,
          suffix: '.cmd',
          entrySuffix: '', // 给主入口添加的后缀
          type: endType,  // 类型
          
          // 控制项
          disableBuiltIn: false,
          disableXtplReplace: false,
          isDev: isDev   // 根据出否是在 dev 状态区分不一样的行为
        };

        // 如果是组件，使用另外一套配置
        if(isComponent){
            depConfig = {
            dependencies: dependencies,
            modBasePath: PATH_COM_BASE, // 也是以 'npm' 开始
            comBasePath: PATH_COM_BASE,
            version: version,
            modName: modName, 
            suffix: '.cmd',
            entrySuffix: '.cmd', // 给主入口添加的后缀
            type: endType,  // 类型
            
            // 控制项
            disableBuiltIn: true,
            disableXtplReplace: true,
            isDev: isDev   // 根据出否是在 dev 状态区分不一样的行为
          }
        };

        // 目前只针对 rax 模块
        return getDepMap(depConfig).then(function(result){
          
          var depsName = 'deps.'+ endType +'.json';
          var modDeps = result.depMap;
          var verJSON = result.verJSON; // 版本信息
          
          console.log('\n ' + modName + ' 的 ' + endType +' 类型依赖表生成完成\n');

          // 生成对应端的 deps.json 文件，放到缓存文件中去
          var depsFile = path.join(pwd, config.cacheDir , util.getDepsFileName(endType));

          // 生成对应端的 version.json 文件，放到缓存文件中去
          var verFile = path.join(pwd, config.cacheDir, util.getVerFileName(endType));
          
          //输出依赖
          fs.writeFileSync(depsFile, JSON.stringify(modDeps));


          // 输出版本
          fs.writeFileSync(verFile, JSON.stringify(verJSON));

          // return depsFile;  // 各个端的 deps.json 所在的路径
          return {
            modName: modName,
            depsFile: depsFile,
            verJSON: verJSON,
            modDeps: modDeps,
            trulyDeps: trulyDeps,
            endType: endType
          }

        });

    })).then(function(depInfos){

      // 有时候 depInfos 数组中有 undefined，需要过滤掉
      debugExtra('=====> 缓存文件路径：',depInfos);

      var filtered = _.filter(depInfos, function(value){
        return isExist(value) && isExist(value.depsFile);
      });

      // 解析各种缓存文件的内容
      return Promise.map(filtered, function(depInfo){
        return _.assign(depInfo, {
          "cacheDirName": path.dirname(depInfo.depsFile)  // 缓存所在的文件夹
        })
      }).then(function(fileObjs){ // 归并多端对象

          debugExtra('=====> 缓存文件对象个数：',fileObjs.length);

          // 按模块文件夹名字归并
          var group = _.groupBy(fileObjs, 'cacheDirName');

          // 挨个迭代不同的组
          _.forOwn(group, function(fileObjArr, cacheDirName){

            var pwd = depsJsonDest || path.dirname(cacheDirName); // 获取模块文件夹位置，优先使用指定的 depsJsonDest

            debugExtra('=====> 最终要输出 deps.json 的位置：', pwd);
            
            // 合并所有的 json
            var finalJSON = {};
            var finalDepsFile = path.join(pwd, 'deps.json');

            // 迭代各个json文件，合并到一起
            _.each(fileObjArr, function(fileObj){
               _.assign(finalJSON, fileObj.modDeps);
            });

            // 一些基本信息只用第一个对象就可以
            var firstObj = fileObjArr[0];

            // 需要将入口依赖更新成真正的依赖内容
            if(!isComponent && firstObj) {

              // 将依赖表更新成真正的依赖，和源码有关；
              var basePath =  [PATH_MOD_BASE, firstObj.modName].join('/');

              var entryNames = [
                basePath+'/index', 
                basePath+'/index.web',
                basePath + '/index.native',
                basePath + '/index.xtpl',
                basePath + '/index.native.xtpl'
              ];
              
              finalJSON = replaceTrulyDeps(finalJSON, entryNames, firstObj.trulyDeps, firstObj.verJSON);
            }

            // 写回到原来的文件中去
            util.writeFileSyncOrNone(finalDepsFile, JSON.stringify(finalJSON)); 
          });

      });

    }).then(function(){
      callback(); // 执行回调
    }).catch(function(err){
      console.log('生成依赖表有误：',err);
      callback(); // 执行回调
    });


  }.bind(this))
};
